/* preprocessor macros for simulating pseudo assembly-code in C */
/* $Id: pseudolib.h,v 1.11 2000/07/20 19:40:04 selinger Exp $ */

#include <stdio.h>
#include <stdlib.h>

/* ====================================================================== */
/* some global declarations */

typedef void* word;
typedef unsigned char byte;

/* simulate a stack */

#ifndef STACKSIZE
#define STACKSIZE 5000
#endif /* STACKSIZE */

word stack[STACKSIZE];
word maxsp = (word)(stack+STACKSIZE-1);

#define SS ((word*)maxsp-(word*)SP)  /* stack size */

word SP, C, A, V, R;  /* special registers */

int jump;        /* holds address jumped to */
int cmp1, cmp2;  /* hold state for comparison */

char *compiled();

/* a dumb main program */
main() {
  char *result = compiled();
  printf("%s\n", result);
  return 0;
}

#define ERR(s)     {fprintf(stderr, "Error: " s "\n"); exit(1);}
#define ERR2(s,d)  {fprintf(stderr, "Error: " s "\n", d); exit(1);}

/* auxiliary function: convert integer to string */
char *int2str (word n) {
  char *str = (char *) malloc (30);
  if (str == NULL) 
    ERR("out of memory");
  sprintf (str, "%d", n);
  return str;
}

/* ====================================================================== */
/* definition of pseudo-lvalues */
#define MEM(A,i)   *((word *)A+i)

/* ---------------------------------------------------------------------- */
/* definition of pseudo-opcodes */

#define LABEL(x)   case (x):

#define LOAD(x,y)  (x) = (word)(y);
#define EXIT(s)     return s;
#define EXITINT(n)  return int2str(n);
#define EXITBOOL(b) return (b) ? "true" : "false";
#define EXITOES(s)     if (SS<=0) return s;
#define EXITOESINT(n)  if (SS<=0) return int2str(n);
#define EXITOESBOOL(b) if (SS<=0) return (b) ? "true" : "false";
#define JUMP(x)    jump = (int)(x); break;

#define ADD(x,y)   (x) += (int)(y);
#define SUB(x,y)   (x) -= (int)(y);
#define MULT(x,y)  (x) = (word)( (int)(x) * (int)(y) );

#define CMP(x,y)   cmp1 = (int)(x); cmp2 = (int)(y);
#define BE(x)      if (cmp1 == cmp2) {jump = (int)(x); break;}
#define BG(x)      if (cmp1 >  cmp2) {jump = (int)(x); break;}
#define BL(x)      if (cmp1 <  cmp2) {jump = (int)(x); break;}
#define BGE(x)     if (cmp1 >= cmp2) {jump = (int)(x); break;}
#define BLE(x)     if (cmp1 <= cmp2) {jump = (int)(x); break;}
#define BNE(x)     if (cmp1 != cmp2) {jump = (int)(x); break;}

#define CALL0(f)             f();
#define CALL1(f,x)           f(x);
#define CALL2(f,x,y)         f(x,y);
#define CALL3(f,x,y,z)       f(x,y,z);
#define CALL4(f,x,y,z,w)     f(x,y,z,w);
#define CALL5(f,x,y,z,w,v)   f(x,y,z,w,v);

#define PRINTINT(x)          printf("%d\n", (x));
#define PRINTBOOL(x)         printf("%s\n", (x)?"true":"false");
#define READINT(x)           printf("Integer: "); scanf("%d", &(x));
#define READBOOL(x)          printf("Boolean: "); \
                             {char sacc[100]; scanf("%s", sacc); \
			      (x) = (word)!strcmp(sacc,"true");}

#ifdef UNSAFE /* in UNSAFE mode omit checks for error conditions */

#define DIV(x,y)   (x) = (word)( (int)(x) / (int)(y) );
#define MOD(x,y)   (x) = (word)( (int)(x) % (int)(y) );
#define POP(x)     (x) = *((word*)SP); SP=(word)((word*)SP+1);
#define PUSH(x)    SP=(word)((word*)SP-1); *((word*)SP) = (word)(x);
#define ALLOC(x,n) (x) = malloc ((n) * sizeof(word));

#else

#define DIV(x,y)   if ((int)(y)==0) ERR("division by zero"); \
                   (x) = (word)( (int)(x) / (int)(y) );
#define MOD(x,y)   if ((int)(y)==0) ERR("modulo zero"); \
                   (x) = (word)( (int)(x) % (int)(y) );
#define POP(x)     if (SS<=0) ERR("stack underflow"); \
                   (x) = *((word*)SP); SP=(word)((word*)SP+1);
#define PUSH(x)    if (SS>=STACKSIZE) \
                     ERR2("stack overflow (STACKSIZE=%d)", STACKSIZE); \
                   SP=(word)((word*)SP-1); *((word*)SP) = (word)(x);
#define ALLOC(x,n) (x) = malloc ((n) * sizeof(word)); \
                   if (!(x)) ERR("out of memory");

#endif /* UNSAFE */

#define REGISTERS word     /* each register (incl. last) is followed by "," */
#define ENDREGISTERS dummy1;

#define LABELS enum {      /* each label (including last) is followed by "," */
#define ENDLABELS dummy2};

#define BEGINCODE  char *compiled() {        \
                   SP = maxsp;               \
                   jump = -1;                \
	           while (1) switch (jump) { \
		   case -1:
#define ENDCODE    }}

/* DATA directives must either occur before, or immediately after BEGINCODE */
/* DLABELS must not be declared */
#define DLABEL(x)  word x[1] =
#define DATA(y)    {(word)y};  

#define BEGINDATA     word datasection[] = {
#define WLABEL(l)     }; word l[] = {
#define WDATA(x)      (word)(x),
#define BLABEL(l)     }; byte l[] = {
#define BDATA(x)      (byte)(x),
#define ENDDATA       }; /* end of data section */

#define COMMENT(x)    /* x */
